from pathlib import Path
from typing import Optional

from dialog2rasa.converters.base import BaseConverter
from dialog2rasa.utils.general import camel_to_snake, logger
from dialog2rasa.utils.io import read_json_file, write_to_file


class IntentConverter(BaseConverter):
    def __init__(
        self,
        agent_dir: Path,
        agent_name: str,
        language: str,
        output_file: Optional[str] = None,
    ) -> None:
        super().__init__(agent_dir, agent_name, language, output_file, "nlu")
        self.intents_path = agent_dir / "intents"

    def convert(self) -> None:
        """Converts Dialogflow intents to Rasa NLU format."""
        converted_intents = self._handle_intents()
        write_to_file(self.output_path, converted_intents)
        logger.debug(f"The file '{self.output_path}' has been created.")

    def _handle_intents(self) -> str:
        """Handles conversion of individual Dialogflow intents."""
        converted_intents = 'version: "3.1"\n\nnlu:\n'
        intent_file_stem = f"_usersays_{self.language}"

        for file in sorted(self.intents_path.glob(f"*{intent_file_stem}.json")):
            intent_name = camel_to_snake(file.stem.split(intent_file_stem)[0])
            data = read_json_file(file)
            examples = self._handle_examples(data)
            converted_intents += (
                f"  - intent: {intent_name}\n    examples: |\n{examples}\n"
            )

        return converted_intents

    def _handle_examples(self, data: dict) -> str:
        """Handles conversion of Dialogflow examples to Rasa format."""
        examples = ""
        for d in data:
            text = "".join(
                (
                    f'[{fragment["text"]}]'
                    f'({fragment["meta"].replace("@", "").replace(".", "_")})'
                    if "meta" in fragment
                    else fragment["text"]
                )
                for fragment in d["data"]
            )
            examples += f"      - {text.strip()}\n"
        return examples



from abc import abstractmethod
from pathlib import Path
from typing import Optional


class BaseConverter:
    """
    Base converter to simplify initialization. Main function are:
    1) Manipulate the strings from Dialogflow export files and;
    2) Export the output files following the RASA YAML-format.
    """

    def __init__(
        self,
        agent_dir: Path,
        agent_name: str,
        language: str,
        output_file: Optional[str] = None,
        sub_dir: str = "",
    ) -> None:
        self.agent_dir = agent_dir
        self.agent_name = agent_name
        self.language = language
        self.output_dir = agent_dir / "output" / self.language
        self.nlu_folder_path = (
            self.output_dir / "data" / "nlu" if sub_dir == "nlu" else self.output_dir
        )
        self.output_file = output_file if output_file else f"{self.agent_name}.yml"
        self.output_path = self.nlu_folder_path / self.output_file

    @abstractmethod
    def convert(self) -> None: ...



import logging
import re
from logging import INFO, Logger, StreamHandler


def setup_logger() -> Logger:
    """Sets up the logger configuration."""
    # Format string including level, module, and function name
    format_str = "%(levelname)s - %(module)s - %(funcName)s - %(message)s"

    # Create and configure logger
    logger = logging.getLogger()
    logger.setLevel(INFO)

    # Create console handler and set level to info
    ch = StreamHandler()
    ch.setLevel(INFO)

    # Create formatter and add it to the handlers
    formatter = logging.Formatter(format_str)
    ch.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(ch)
    return logger


logger = setup_logger()


def camel_to_snake(s: str) -> str:
    """Converts input string from CamelCase to snake_case."""
    return (
        re.sub(r"(?<!^)(?=[A-Z])", "_", s)  # insert underscores before capital letters
        .replace(" ", "_")  # fix non non-standard CamelCase inputs
        .replace("__", "_")  # fix double underscores (from previous replace)
        .lower()  # lowercase for true snake_case
    )



from pathlib import Path
import sys
from typing import Optional

from dialog2rasa.converters.base import BaseConverter
from dialog2rasa.converters.manager import get_converter
from dialog2rasa.utils.general import logger
from dialog2rasa.utils.io import reset_directory


class DialogflowToRasaConverter(BaseConverter):
    """Converts Dialogflow agent files (.zip export) to Rasa format."""

    def __init__(
        self,
        agent_dir: Path,
        language: str,
        output_file: Optional[str] = None,
    ) -> None:
        """Initializes converter settings."""
        super().__init__(agent_dir, agent_dir.stem, language, output_file, "nlu")
        if not self._check_language_files_existence():
            logger.error(
                f"Language code '{self.language}' files not found "
                "in intents or entities directories. Please check if "
                "the language was correctly input"
            )
            sys.exit(1)

    def _check_language_files_existence(self) -> bool:
        """Checks if language-specific files exist in the agent directories."""
        intents_path = self.agent_dir / "intents"
        entities_path = self.agent_dir / "entities"
        language_files_exist = any(
            file.name.endswith(f"_{self.language}.json")
            for file in intents_path.glob("*.json")
        ) and any(
            file.name.endswith(f"_entries_{self.language}.json")
            for file in entities_path.glob("*.json")
        )
        return language_files_exist

    def _initialize_converters(self) -> None:
        """Initializes all required converters."""
        converter_types = ["intent", "utterance", "entity"]
        self.converters: dict[str, BaseConverter] = {}
        for converter_type in converter_types:
            self.converters[converter_type] = get_converter(
                converter_type, self.agent_dir, self.agent_name, self.language
            )

    def convert_all(self) -> None:
        """Converts all Dialogflow files to Rasa format."""
        logger.debug("Starting conversion...")
        reset_directory(self.output_dir, "data/nlu/lookup")

        self._initialize_converters()

        for _, converter in self.converters.items():
            converter.convert()

        logger.info(
            "Conversion completed. "
            f"The output files can be found in '{self.output_dir}'"
        )



from dialog2rasa.converters.base import BaseConverter
from dialog2rasa.converters.entity import EntityConverter
from dialog2rasa.converters.intent import IntentConverter
from dialog2rasa.converters.utterance import UtteranceConverter


def get_converter(type_: str, *args, **kwargs) -> BaseConverter:
    converters = {
        "intent": IntentConverter,
        "utterance": UtteranceConverter,
        "entity": EntityConverter,
    }
    return converters[type_](*args, **kwargs)



from pathlib import Path
from typing import Optional

from dialog2rasa.converters.base import BaseConverter
from dialog2rasa.utils.general import camel_to_snake, logger
from dialog2rasa.utils.io import read_json_file, write_to_file


class UtteranceConverter(BaseConverter):
    def __init__(
        self,
        agent_dir: Path,
        agent_name: str,
        language: str,
        output_file: Optional[str] = "domain.yml",
    ) -> None:
        super().__init__(agent_dir, agent_name, language, output_file)

    def convert(self) -> None:
        """Converts Dialogflow utterances to Rasa domain format."""
        responses_folder_path = self.agent_dir / "intents"
        converted_responses = self._handle_responses(responses_folder_path)
        write_to_file(self.output_path, converted_responses)
        logger.debug(f"The file '{self.output_path}' has been created.")

    def _handle_responses(self, responses_folder_path: Path) -> str:
        """Handles conversion of Dialogflow responses to Rasa format."""
        converted_responses = "responses:\n"
        for file in sorted(responses_folder_path.iterdir()):
            if not file.name.endswith(f"usersays_{self.language}.json"):
                intent_name = camel_to_snake(file.stem)
                data = read_json_file(file)
                for response in data.get("responses", []):
                    converted_responses += self._handle_utter_message(
                        intent_name, response
                    )
        return converted_responses

    def _handle_utter_message(self, intent_name: str, response: dict) -> str:
        """Handles conversion of individual Dialogflow utter messages."""
        converted_utter = ""
        for message in response.get("messages", []):
            if message.get("lang") == self.language:
                if "speech" in message:
                    converted_utter += f"  utter_{intent_name}:\n"
                    for s in message["speech"]:
                        converted_utter += f'    - text: "{s}"\n'
                    converted_utter += "\n"
        return converted_utter



from pathlib import Path
from typing import Optional

from dialog2rasa.converters.base import BaseConverter
from dialog2rasa.utils.general import camel_to_snake, logger
from dialog2rasa.utils.io import read_json_file, write_dict_files


class EntityConverter(BaseConverter):
    def __init__(
        self,
        agent_dir: Path,
        agent_name: str,
        language: str,
        output_file: Optional[str] = None,
    ) -> None:
        super().__init__(agent_dir, agent_name, language, output_file, "nlu")
        self.lookup_path = self.nlu_folder_path / "lookup"
        self.entities_path = self.agent_dir / "entities"
        self.domain_file_path = self.output_dir / "domain.yml"

    def convert(self) -> None:
        """Processes and converts Dialogflow entities to Rasa format."""
        entity_contents = self._handle_entities()
        for entity_dict in entity_contents:
            write_dict_files(entity_dict)
        self._handle_entities_as_slots()
        logger.debug(
            f"The entity files have been created in dir '{self.nlu_folder_path}'."
        )

    def _handle_entities(self) -> tuple:
        synonym_content, lookup_content, compound_content = {}, {}, {}

        for entity_file in self.entities_path.glob(f"*_entries_{self.language}.json"):
            entity_name = camel_to_snake(entity_file.stem).replace(
                f"_entries_{self.language}", ""
            )
            entries = read_json_file(entity_file)

            for entry in entries:
                if any("@" in syn for syn in entry["synonyms"]):
                    # @ refers to compound entities in dialogflow
                    compound_file_path = (
                        self.nlu_folder_path / f"__compound__{entity_name}.yml"
                    )
                    if compound_file_path not in compound_content:
                        compound_content[compound_file_path] = (
                            self._init_compound_file_content()
                        )
                        logger.warning(
                            "Manual adaptation needed for compound "
                            f"entity '{entity_name}' in Rasa. "
                            f"See file: '__compound__{entity_name}.yml'."
                        )
                    self._update_content(
                        compound_content,
                        compound_file_path,
                        self._handle_compounds(entry),
                    )

                elif len(entry["synonyms"]) > 1:
                    # multiple synonyms are considered also synonyms in RASA
                    synonyms_file_path = self.nlu_folder_path / f"{self.agent_name}.yml"
                    self._update_content(
                        synonym_content,
                        synonyms_file_path,
                        self._handle_synonyms(entry),
                    )

                else:
                    # single synonyms are accumulated in lookup tables
                    lookup_file_path = self.lookup_path / f"{entity_name}.txt"
                    self._update_content(
                        lookup_content,
                        lookup_file_path,
                        self._handle_lookup(entry),
                    )

        return synonym_content, lookup_content, compound_content

    def _update_content(self, content_dict: dict, file_path: Path, new_content: str):
        if file_path not in content_dict:
            content_dict[file_path] = ""
        content_dict[file_path] += new_content

    def _handle_synonyms(self, entry: dict) -> str:
        """Returns Rasa format string for Dialogflow synonyms."""
        synonym = entry["value"]
        examples = "\n".join(f"      - {syn}" for syn in entry["synonyms"])
        return f"  - synonym: {synonym}\n    examples: |\n{examples}\n\n"

    def _handle_lookup(self, entry: dict) -> str:
        """Returns Rasa format string for Dialogflow lookup tables."""
        return "\n".join(entry["synonyms"]) + "\n"

    def _init_compound_file_content(self) -> str:
        """Returns the initial content for a new compound entity file."""
        header_content = "# Compound entity: Manual adaptation needed for Rasa\n"
        header_content += 'version: "3.1"\n\nnlu:\n'
        return header_content

    def _handle_compounds(self, entry: dict) -> str:
        """
        Returns Rasa format string for pseudo-compound entities from Dialogflow.
        No need to check for new files here, handled in convert method.
        """
        synonym = entry["value"]
        examples = "\n".join(f"      - {syn}" for syn in entry["synonyms"])
        return f"  - synonym: {synonym}\n    examples: |\n{examples}\n\n"

    def _handle_entities_as_slots(self):
        """Appends entities as slots to the Rasa domain file."""
        if not self.domain_file_path.exists():
            logger.error(f"Domain file {self.domain_file_path} not found.")
            return

        with self.domain_file_path.open("a") as domain_file:

            entity_names = sorted(
                set(
                    [
                        x.stem.split("_entries")[0]
                        for x in self.entities_path.glob(
                            f"*_entries_{self.language}.json"
                        )
                    ]
                )
            )
            entities_str = "\n  - ".join(entity_names)
            domain_file.write(
                "# TODO: Review assumption of Dialogflow entities as slots and entities. Confirm the types and mappings.\nentities:"
                f"\n  - {entities_str}"
                "\n\nslots:\n"
            )
            for entity_name in entity_names:
                domain_file.write(
                    f"  {entity_name}:\n    type: text\n    "
                    "influence_conversation: false \n    mappings:\n    "
                    f"- type: from_entity\n      entity: {entity_name}\n\n"
                )

        logger.warning(
            "Entities have been added as slots to the domain file. "
            "Please review slot types and mappings."
        )



import json
import shutil
from pathlib import Path

from dialog2rasa.utils.general import logger


def reset_directory(dir_path: Path, deepest_subdir: str) -> None:
    """
    Deletes all files and folders in the specified directory recursively,
    then creates the specified deepest subdirectory.
    """
    if dir_path.exists():
        logger.warning(f"Clearing all contents from '{dir_path}'...")
        shutil.rmtree(dir_path)
        logger.debug("Output directory cleared.")
    (dir_path / deepest_subdir).mkdir(parents=True, exist_ok=True)


def read_json_file(file_path: Path) -> dict:
    """Reads and returns JSON data from a file."""
    with Path(file_path).open("r", encoding="utf-8") as file:
        return json.load(file)


def write_to_file(file_path: Path, content: str, mode: str = "w") -> None:
    """Writes given content to a file."""
    with Path(file_path).open(mode, encoding="utf-8") as file:
        file.write(content)


def write_dict_files(content_dict: dict) -> None:
    """Writes given content to corresponding files."""
    for file_path, content in content_dict.items():
        write_to_file(file_path, content, mode="a")



import argparse
from pathlib import Path

from dialog2rasa.converters.core import DialogflowToRasaConverter


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Transforms a Dialogflow agent into Rasa format. "
        "The result is saved in /output/[LANGUAGE_CODE], where [LANGUAGE_CODE] "
        "is replaced with the actual language code (e.g., 'en', 'de'), inside "
        "the Dialogflow agent's directory."
    )
    parser.add_argument(
        "--path",
        "-p",
        required=True,
        help="Path to the Dialogflow agent's extracted/unzipped folder.",
    )
    parser.add_argument(
        "--language",
        "-l",
        default="de",
        help="Language code (e.g., 'en' for English) of the Dialogflow agent. "
        "Defaults to 'de' (German).",
    )

    args = parser.parse_args()

    converter = DialogflowToRasaConverter(
        agent_dir=Path(args.path), language=args.language
    )
    converter.convert_all()


if __name__ == "__main__":
    main()
